@@ -161,6 +161,8 @@ class NeuralInference(ABC):
             epochs=[],
             best_validation_log_probs=[],
             validation_log_probs=[],
+            train_log_probs=[],
+            epoch_durations_sec=[],
         )
 
     def __call__(self, unused_args):
@@ -434,13 +436,6 @@ class NeuralInference(ABC):
                 "but network has not yet fully converged. Consider increasing it."
             )
 
-    @staticmethod
-    def _assert_all_finite(quantity: Tensor, description: str = "tensor") -> None:
-        """Raise if tensor quantity contains any NaN or Inf element."""
-
-        msg = f"NaN/Inf present in {description}."
-        assert torch.isfinite(quantity).all(), msg
-
     def _summarize(
         self,
         round_: int,
@@ -501,6 +496,20 @@ class NeuralInference(ABC):
                 global_step=offset + i,
             )
 
+        for i, tlp in enumerate(self._summary["train_log_probs"][offset:]):
+            self._summary_writer.add_scalar(
+                tag="train_log_probs_across_rounds",
+                scalar_value=tlp,
+                global_step=offset + i,
+            )
+
+        for i, eds in enumerate(self._summary["epoch_durations_sec"][offset:]):
+            self._summary_writer.add_scalar(
+                tag="epoch_durations_sec_across_rounds",
+                scalar_value=eds,
+                global_step=offset + i,
+            )
+
         self._summary_writer.flush()
 
     @property
