@@ -25,13 +25,14 @@ from sbi.mcmc import (
     sir,
 )
 from sbi.types import Array, Shape
-from sbi.utils.sbiutils import check_if_boxuniform
+from sbi.utils.sbiutils import check_dist_class, check_warn_and_setstate
 from sbi.utils.torchutils import (
+    BoxUniform,
     ScalarFloat,
     atleast_2d_float32_tensor,
     ensure_theta_batched,
 )
-from sbi.utils.user_input_checks import process_x
+from sbi.utils.user_input_checks import check_for_possibly_batched_x_shape, process_x
 
 
 class NeuralPosterior(ABC):
@@ -58,7 +59,7 @@ class NeuralPosterior(ABC):
             method_family: One of snpe, snl, snre_a or snre_b.
             neural_net: A classifier for SNRE, a density estimator for SNPE and SNL.
             prior: Prior distribution with `.log_prob()` and `.sample()`.
-            x_shape: Shape of a single simulator output.
+            x_shape: Shape of the simulator data.
             mcmc_method: Method used for MCMC sampling, one of `slice_np`, `slice`,
                 `hmc`, `nuts`. Currently defaults to `slice_np` for a custom numpy
                 implementation of slice sampling; select `hmc`, `nuts` or `slice` for
@@ -76,7 +77,7 @@ class NeuralPosterior(ABC):
         if method_family in ("snpe", "snle", "snre_a", "snre_b"):
             self._method_family = method_family
         else:
-            raise ValueError("Method family unsupported.")
+            raise ValueError(f"Method family '{method_family}' unsupported.")
 
         self.net = neural_net
 
@@ -88,8 +89,15 @@ class NeuralPosterior(ABC):
         self._num_trained_rounds = 0
         self._prior = prior
         self._x = None
+        self._num_iid_trials = None
         self._x_shape = x_shape
         self._device = device
+        # Methods capable of handling iid xo.
+        self._iid_methods = ["snle", "snre_a", "snre_b"]
+        self._allow_iid_x = method_family in self._iid_methods
+
+        if not self._allow_iid_x:
+            check_for_possibly_batched_x_shape(self._x_shape)
 
     @property
     def default_x(self) -> Optional[Tensor]:
@@ -121,8 +129,8 @@ class NeuralPosterior(ABC):
         Returns:
             `NeuralPosterior` that will use a default `x` when not explicitly passed.
         """
-        processed_x = process_x(x, self._x_shape)
-        self._x = processed_x
+        self._x = process_x(x, self._x_shape, allow_iid_x=self._allow_iid_x)
+        self._num_iid_trials = self._x.shape[0]
 
         return self
 
@@ -182,10 +190,7 @@ class NeuralPosterior(ABC):
 
     @abstractmethod
     def log_prob(
-        self,
-        theta: Tensor,
-        x: Optional[Tensor] = None,
-        track_gradients: bool = False,
+        self, theta: Tensor, x: Optional[Tensor] = None, track_gradients: bool = False
     ) -> Tensor:
         """See child classes for docstring."""
         pass
@@ -240,15 +245,15 @@ class NeuralPosterior(ABC):
         return self
 
     def _prepare_theta_and_x_for_log_prob_(
-        self,
-        theta: Tensor,
-        x: Optional[Tensor] = None,
+        self, theta: Tensor, x: Optional[Tensor] = None
     ) -> Tuple[Tensor, Tensor]:
         r"""Returns $\theta$ and $x$ in shape that can be used by posterior.log_prob().
 
         Checks shapes of $\theta$ and $x$ and then repeats $x$ as often as there were
         batch elements in $\theta$.
 
+        Moves $\theta$ and $x$ to the device of the neural net.
+
         Args:
             theta: Parameters $\theta$.
             x: Conditioning context for posterior $p(\theta|x)$. If not provided, fall
@@ -264,14 +269,10 @@ class NeuralPosterior(ABC):
 
         # Select and check x to condition on.
         x = atleast_2d_float32_tensor(self._x_else_default_x(x))
-        self._ensure_single_x(x)
+        if not self._allow_iid_x:
+            self._ensure_single_x(x)
         self._ensure_x_consistent_with_default_x(x)
 
-        # Repeat `x` in case of evaluation on multiple `theta`. This is needed below in
-        # when calling nflows in order to have matching shapes of theta and context x
-        # at neural network evaluation time.
-        x = self._match_x_with_theta_batch_shape(x, theta)
-
         return theta, x
 
     def _prepare_for_sample(
@@ -307,7 +308,8 @@ class NeuralPosterior(ABC):
         """
 
         x = atleast_2d_float32_tensor(self._x_else_default_x(x))
-        self._ensure_single_x(x)
+        if not self._allow_iid_x:
+            self._ensure_single_x(x)
         self._ensure_x_consistent_with_default_x(x)
         num_samples = torch.Size(sample_shape).numel()
 
@@ -317,9 +319,7 @@ class NeuralPosterior(ABC):
         )
 
         # Move x to current device.
-        x = x.to(self._device)
-
-        return x, num_samples, mcmc_method, mcmc_parameters
+        return x.to(self._device), num_samples, mcmc_method, mcmc_parameters
 
     def _sample_posterior_mcmc(
         self,
@@ -659,7 +659,7 @@ class NeuralPosterior(ABC):
 
         # If the prior is `BoxUniform`, define a transformation to optimize in
         # unbounded space.
-        is_boxuniform, boxuniform = check_if_boxuniform(self._prior)
+        is_boxuniform, boxuniform = check_dist_class(self._prior, BoxUniform)
         if is_boxuniform:
 
             def tf_inv(theta_t):
@@ -763,14 +763,14 @@ class NeuralPosterior(ABC):
                             best_theta_overall = best_theta_iter.detach().clone()
                             best_log_prob_overall = best_log_prob_iter.detach().clone()
 
-                    print(
-                        f"Optimizing MAP estimate. Iterations: "
-                        f"{iter_+1} / {num_iter}.    "
-                        f"Performance in iteration "
-                        f"{divmod(iter_+1, save_best_every)[0] * save_best_every}: "
-                        f"{best_log_prob_iter.item():.2f} (= unnormalized log-prob)",
-                        end="\r",
-                    )
+                    if show_progress_bars:
+                        print(
+                            f"""Optimizing MAP estimate. Iterations: {iter_+1} /
+                            {num_iter}. Performance in iteration
+                            {divmod(iter_+1, save_best_every)[0] * save_best_every}:
+                            {best_log_prob_iter.item():.2f} (= unnormalized log-prob""",
+                            end="\r",
+                        )
                     self.map_ = tf_inv(best_theta_overall)
 
                 iter_ += 1
@@ -850,10 +850,11 @@ class NeuralPosterior(ABC):
             raise ValueError(
                 """The `x` passed to condition the posterior for evaluation or sampling
                 has an inferred batch shape larger than one. This is not supported in
-                sbi for reasons depending on the scenario:
+                some sbi methods for reasons depending on the scenario:
 
                     - in case you want to evaluate or sample conditioned on several xs
-                    e.g., (p(theta | [x1, x2, x3])), this is not supported yet in sbi.
+                    e.g., (p(theta | [x1, x2, x3])), this is not supported yet except
+                    when using likelihood based SNLE.
 
                     - in case you trained with a single round to do amortized inference
                     and now you want to evaluate or sample a given theta conditioned on
@@ -871,29 +872,46 @@ class NeuralPosterior(ABC):
             )
 
     @staticmethod
-    def _match_x_with_theta_batch_shape(x: Tensor, theta: Tensor) -> Tensor:
-        """Return `x` with batch shape matched to that of `theta`.
+    def _match_theta_and_x_batch_shapes(
+        theta: Tensor, x: Tensor
+    ) -> Tuple[Tensor, Tensor]:
+        r"""Return $\theta$ and `x` with batch shape matched to each other.
+
+        When `x` is just a single observation it is repeated for all entries in the
+        batch of $\theta$s. When there is a batch of multiple `x`, i.e., iid `x`, then
+        individual `x` are repeated in the pattern AABBCC and individual $\theta$ are
+        repeated in the pattern ABCABC to cover all combinations.
 
         This is needed in nflows in order to have matching shapes of theta and context
         `x` when evaluating the neural network.
+
+        Args:
+            x: (a batch of iid) data
+            theta: a batch of parameters
+
+        Returns:
+            theta: with shape (theta_batch_size * x_batch_size, *theta_shape)
+            x: with shape (theta_batch_size * x_batch_size, *x_shape)
         """
 
         # Theta and x are ensured to have a batch dim, get the shape.
-        theta_batch_size, *_ = theta.shape
+        theta_batch_size, *theta_shape = theta.shape
         x_batch_size, *x_shape = x.shape
 
-        assert x_batch_size == 1, "Batch size 1 should be enforced by caller."
-        if theta_batch_size > x_batch_size:
-            x_matched = x.expand(theta_batch_size, *x_shape)
+        # Repeat iid trials as AABBCC.
+        x_repeated = x.repeat_interleave(theta_batch_size, dim=0)
+        # Repeat theta as ABCABC.
+        theta_repeated = theta.repeat(x_batch_size, 1)
 
-            # Double check.
-            x_matched_batch_size, *x_matched_shape = x_matched.shape
-            assert x_matched_batch_size == theta_batch_size
-            assert x_matched_shape == x_shape
-        else:
-            x_matched = x
+        # Double check: batch size for log prob evaluation must match.
+        assert x_repeated.shape == torch.Size(
+            [theta_batch_size * x_batch_size, *x_shape]
+        )
+        assert theta_repeated.shape == torch.Size(
+            [theta_batch_size * x_batch_size, *theta_shape]
+        )
 
-        return x_matched
+        return theta_repeated, x_repeated
 
     def _get_net_name(self) -> str:
         """
@@ -962,16 +980,48 @@ class NeuralPosterior(ABC):
         """
         Sets the state when being loaded from pickle.
 
+        For developers: for any new attribute added to `NeuralPosterior`, we have to
+        add an entry here using `check_warn_and_setstate()`.
+
         Args:
             state_dict: State to be restored.
         """
-        if "_device" not in state_dict.keys():
-            state_dict["_device"] = "cpu"
-            warn(
+
+        # In the beginning, the warning message is empty.
+        warning_msg = ""
+
+        state_dict, warning_msg = check_warn_and_setstate(
+            state_dict, "_device", "cpu", warning_msg
+        )
+
+        if state_dict["_x"] is not None:
+            state_dict, warning_msg = check_warn_and_setstate(
+                state_dict, "_num_iid_trials", state_dict["_x"].shape[0], warning_msg
+            )
+        else:
+            state_dict, warning_msg = check_warn_and_setstate(
+                state_dict, "_num_iid_trials", None, warning_msg
+            )
+
+        state_dict, warning_msg = check_warn_and_setstate(
+            state_dict, "_iid_methods", ["snle", "snre_a", "snre_b"], warning_msg
+        )
+
+        state_dict, warning_msg = check_warn_and_setstate(
+            state_dict,
+            "_allow_iid_x",
+            state_dict["_method_family"] in state_dict["_iid_methods"],
+            warning_msg,
+        )
+
+        if warning_msg:
+            warning_description = (
                 "You had saved the posterior under an older version of `sbi`. To make "
                 "the loaded version comply with the version you are using right now, "
-                "we had to set `self._device = 'cpu'`"
+                "we had to set the following attributes:"
             )
+            warn(warning_description + warning_msg)
+
         self.__dict__ = state_dict
 
 
@@ -1003,13 +1053,7 @@ class ConditionalPotentialFunctionProvider:
         self.condition = ensure_theta_batched(condition)
         self.dims_to_sample = dims_to_sample
 
-    def __call__(
-        self,
-        prior,
-        net: nn.Module,
-        x: Tensor,
-        mcmc_method: str,
-    ) -> Callable:
+    def __call__(self, prior, net: nn.Module, x: Tensor, mcmc_method: str) -> Callable:
         """Return potential function.
 
         Switch on numpy or pyro potential function based on `mcmc_method`.
@@ -1076,11 +1120,7 @@ class RestrictedPriorForConditional:
     this class.
     """
 
-    def __init__(
-        self,
-        full_prior: Any,
-        dims_to_sample: List[int],
-    ):
+    def __init__(self, full_prior: Any, dims_to_sample: List[int]):
         self.full_prior = full_prior
         self.dims_to_sample = dims_to_sample
 
