@@ -1,6 +1,7 @@
 # This file is part of sbi, a toolkit for simulation-based inference. sbi is licensed
 # under the Affero General Public License v3, see <https://www.gnu.org/licenses/>.
 
+from functools import partial
 from typing import Any, Callable, Dict, List, Optional, Union
 
 import numpy as np
@@ -10,11 +11,11 @@ from torch import Tensor, log, nn
 from sbi import utils as utils
 from sbi.inference.posteriors.base_posterior import NeuralPosterior
 from sbi.types import ScalarFloat, Shape
-from sbi.utils import del_entries
+from sbi.utils import del_entries, within_support
 from sbi.utils.torchutils import (
+    atleast_2d,
     batched_first_of_batch,
     ensure_theta_batched,
-    ensure_x_batched,
 )
 
 
@@ -193,19 +194,20 @@ class DirectPosterior(NeuralPosterior):
         self.net.eval()
 
         theta, x = self._prepare_theta_and_x_for_log_prob_(theta, x)
+        theta_repeated, x_repeated = self._match_theta_and_x_batch_shapes(theta, x)
 
         with torch.set_grad_enabled(track_gradients):
 
             # Evaluate on device, move back to cpu for comparison with prior.
             unnorm_log_prob = self.net.log_prob(
-                theta.to(self._device), x.to(self._device)
+                theta_repeated.to(self._device), x_repeated.to(self._device)
             ).cpu()
 
             # Force probability to be zero outside prior support.
-            is_prior_finite = torch.isfinite(self._prior.log_prob(theta))
+            in_prior_support = within_support(self._prior, theta)
 
             masked_log_prob = torch.where(
-                is_prior_finite,
+                in_prior_support,
                 unnorm_log_prob,
                 torch.tensor(float("-inf"), dtype=torch.float32),
             )
@@ -521,10 +523,13 @@ class PotentialFunctionProvider:
         """
         self.posterior_nn = posterior_nn
         self.prior = prior
-        self.x = x
+        self.device = next(posterior_nn.parameters()).device
+        self.x = atleast_2d(x).to(self.device)
 
-        if mcmc_method in ("slice", "hmc", "nuts"):
-            return self.pyro_potential
+        if mcmc_method == "slice":
+            return partial(self.pyro_potential, track_gradients=False)
+        elif mcmc_method in ("hmc", "nuts"):
+            return partial(self.pyro_potential, track_gradients=True)
         else:
             return self.np_potential
 
@@ -541,23 +546,22 @@ class PotentialFunctionProvider:
         theta = ensure_theta_batched(theta)
         num_batch = theta.shape[0]
 
-        x_batched = ensure_x_batched(self.x)
         # Repeat x over batch dim to match theta batch, accounting for multi-D x.
-        x_repeated = x_batched.repeat(
-            num_batch, *(1 for _ in range(x_batched.ndim - 1))
-        )
+        x_repeated = self.x.repeat(num_batch, *(1 for _ in range(self.x.ndim - 1)))
 
         with torch.set_grad_enabled(False):
             target_log_prob = self.posterior_nn.log_prob(
-                inputs=theta.to(self.x.device),
+                inputs=theta.to(self.device),
                 context=x_repeated,
             )
-            is_within_prior = torch.isfinite(self.prior.log_prob(theta))
-            target_log_prob[~is_within_prior] = -float("Inf")
+            in_prior_support = within_support(self.prior, theta)
+            target_log_prob[~in_prior_support] = -float("Inf")
 
         return target_log_prob
 
-    def pyro_potential(self, theta: Dict[str, Tensor]) -> Tensor:
+    def pyro_potential(
+        self, theta: Dict[str, Tensor], track_gradients: bool = False
+    ) -> Tensor:
         r"""Return posterior log prob. of theta $p(\theta|x)$, -inf where outside prior.
 
         Args:
@@ -569,13 +573,18 @@ class PotentialFunctionProvider:
 
         theta = next(iter(theta.values()))
 
-        # Notice opposite sign to numpy.
-        # Move theta to device for evaluation.
-        log_prob_posterior = -self.posterior_nn.log_prob(
-            inputs=theta.to(self.x.device), context=self.x
-        ).cpu()
-        log_prob_prior = self.prior.log_prob(theta)
+        with torch.set_grad_enabled(track_gradients):
+            # Notice opposite sign to numpy.
+            # Move theta to device for evaluation.
+            log_prob_posterior = -self.posterior_nn.log_prob(
+                inputs=theta.to(self.device),
+                context=self.x,
+            ).cpu()
 
-        within_prior = torch.isfinite(log_prob_prior)
+        in_prior_support = within_support(self.prior, theta)
 
-        return torch.where(within_prior, log_prob_posterior, log_prob_prior)
+        return torch.where(
+            in_prior_support,
+            log_prob_posterior,
+            float("-inf") * torch.ones_like(log_prob_posterior),
+        )
