@@ -1,6 +1,7 @@
 # This file is part of sbi, a toolkit for simulation-based inference. sbi is licensed
 # under the Affero General Public License v3, see <https://www.gnu.org/licenses/>.
 
+from functools import partial
 from typing import Any, Callable, Dict, List, Optional, Union
 from warnings import warn
 
@@ -11,7 +12,7 @@ from torch import Tensor, nn
 from sbi.inference.posteriors.base_posterior import NeuralPosterior
 from sbi.types import Shape
 from sbi.utils import del_entries
-from sbi.utils.torchutils import ScalarFloat, ensure_theta_batched, ensure_x_batched
+from sbi.utils.torchutils import ScalarFloat, atleast_2d, ensure_theta_batched
 
 
 class LikelihoodBasedPosterior(NeuralPosterior):
@@ -39,7 +40,12 @@ class LikelihoodBasedPosterior(NeuralPosterior):
             method_family: One of snpe, snl, snre_a or snre_b.
             neural_net: A classifier for SNRE, a density estimator for SNPE and SNL.
             prior: Prior distribution with `.log_prob()` and `.sample()`.
-            x_shape: Shape of a single simulator output.
+            x_shape: Shape of the simulated data. It can differ from the
+                observed data the posterior is conditioned on later in the batch
+                dimension. If it differs, the additional entries are interpreted as
+                independent and identically distributed data / trials. I.e., the data is
+                assumed to be generated based on the same (unknown) model parameters or
+                experimental condations.
             mcmc_method: Method used for MCMC sampling, one of `slice_np`, `slice`,
                 `hmc`, `nuts`. Currently defaults to `slice_np` for a custom numpy
                 implementation of slice sampling; select `hmc`, `nuts` or `slice` for
@@ -59,15 +65,12 @@ class LikelihoodBasedPosterior(NeuralPosterior):
         super().__init__(**kwargs)
 
         self._purpose = (
-            f"It provides MCMC to .sample() from the posterior and "
-            f"can evaluate the _unnormalized_ posterior density with .log_prob()."
+            "It provides MCMC to .sample() from the posterior and "
+            "can evaluate the _unnormalized_ posterior density with .log_prob()."
         )
 
     def log_prob(
-        self,
-        theta: Tensor,
-        x: Optional[Tensor] = None,
-        track_gradients: bool = False,
+        self, theta: Tensor, x: Optional[Tensor] = None, track_gradients: bool = False
     ) -> Tensor:
         r"""
         Returns the log-probability of $p(x|\theta) \cdot p(\theta).$
@@ -92,15 +95,19 @@ class LikelihoodBasedPosterior(NeuralPosterior):
 
         theta, x = self._prepare_theta_and_x_for_log_prob_(theta, x)
 
+        # Calculate likelihood over trials and in one batch.
         warn(
             "The log probability from SNL is only correct up to a normalizing constant."
         )
+        log_likelihood_trial_sum = self._log_likelihoods_over_trials(
+            x=x.to(self._device),
+            theta=theta.to(self._device),
+            net=self.net,
+            track_gradients=track_gradients,
+        )
 
-        with torch.set_grad_enabled(track_gradients):
-            # Evaluate on device, move to cpu for comparison with prior.
-            return self.net.log_prob(
-                x.to(self._device), theta.to(self._device)
-            ).cpu() + self._prior.log_prob(theta)
+        # Move to cpu for comparison with prior.
+        return log_likelihood_trial_sum.cpu() + self._prior.log_prob(theta)
 
     def sample(
         self,
@@ -279,6 +286,55 @@ class LikelihoodBasedPosterior(NeuralPosterior):
             show_progress_bars=show_progress_bars,
         )
 
+    @staticmethod
+    def _log_likelihoods_over_trials(
+        x: Tensor,
+        theta: Tensor,
+        net: nn.Module,
+        track_gradients: bool = False,
+    ) -> Tensor:
+        r"""Return log likelihoods summed over iid trials of `x`.
+
+        Note: `x` can be a batch with batch size larger 1. Batches in `x` are assumed
+        to be iid trials, i.e., data generated based on the same paramters /
+        experimental conditions.
+
+        Repeats `x` and $\theta$ to cover all their combinations of batch entries.
+
+        Args:
+            x: batch of iid data.
+            theta: batch of parameters
+            net: neural net with .log_prob()
+            track_gradients: Whether to track gradients.
+
+        Returns:
+            log_likelihood_trial_sum: log likelihood for each parameter, summed over all
+                batch entries (iid trials) in `x`.
+        """
+
+        # Repeat `x` in case of evaluation on multiple `theta`. This is needed below in
+        # when calling nflows in order to have matching shapes of theta and context x
+        # at neural network evaluation time.
+        theta_repeated, x_repeated = NeuralPosterior._match_theta_and_x_batch_shapes(
+            theta=theta, x=atleast_2d(x)
+        )
+        assert (
+            x_repeated.shape[0] == theta_repeated.shape[0]
+        ), "x and theta must match in batch shape."
+        assert (
+            next(net.parameters()).device == x.device and x.device == theta.device
+        ), f"device mismatch: net, x, theta: {next(net.parameters()).device}, {x.decive}, {theta.device}."
+
+        # Calculate likelihood in one batch.
+        with torch.set_grad_enabled(track_gradients):
+            log_likelihood_trial_batch = net.log_prob(x_repeated, theta_repeated)
+            # Reshape to (x-trials x parameters), sum over trial-log likelihoods.
+            log_likelihood_trial_sum = log_likelihood_trial_batch.reshape(
+                x.shape[0], -1
+            ).sum(0)
+
+        return log_likelihood_trial_sum
+
 
 class PotentialFunctionProvider:
     """
@@ -300,11 +356,7 @@ class PotentialFunctionProvider:
     """
 
     def __call__(
-        self,
-        prior,
-        likelihood_nn: nn.Module,
-        x: Tensor,
-        mcmc_method: str,
+        self, prior, likelihood_nn: nn.Module, x: Tensor, mcmc_method: str
     ) -> Callable:
         r"""Return potential function for posterior $p(\theta|x)$.
 
@@ -313,7 +365,8 @@ class PotentialFunctionProvider:
         Args:
             prior: Prior distribution that can be evaluated.
             likelihood_nn: Neural likelihood estimator that can be evaluated.
-            x: Conditioning variable for posterior $p(\theta|x)$.
+            x: Conditioning variable for posterior $p(\theta|x)$. Can be a batch of iid
+                x.
             mcmc_method: One of `slice_np`, `slice`, `hmc` or `nuts`.
 
         Returns:
@@ -321,13 +374,28 @@ class PotentialFunctionProvider:
         """
         self.likelihood_nn = likelihood_nn
         self.prior = prior
-        self.x = x
+        self.device = next(likelihood_nn.parameters()).device
+        self.x = atleast_2d(x).to(self.device)
 
-        if mcmc_method in ("slice", "hmc", "nuts"):
-            return self.pyro_potential
+        if mcmc_method == "slice":
+            return partial(self.pyro_potential, track_gradients=False)
+        elif mcmc_method in ("hmc", "nuts"):
+            return partial(self.pyro_potential, track_gradients=True)
         else:
             return self.np_potential
 
+    def log_likelihood(self, theta: Tensor, track_gradients: bool = False) -> Tensor:
+        """Return log likelihood of fixed data given a batch of parameters."""
+
+        log_likelihoods = LikelihoodBasedPosterior._log_likelihoods_over_trials(
+            x=self.x,
+            theta=ensure_theta_batched(theta).to(self.device),
+            net=self.likelihood_nn,
+            track_gradients=track_gradients,
+        )
+
+        return log_likelihoods
+
     def np_potential(self, theta: np.array) -> ScalarFloat:
         r"""Return posterior log prob. of theta $p(\theta|x)$"
 
@@ -338,20 +406,13 @@ class PotentialFunctionProvider:
             Posterior log probability of the theta, $-\infty$ if impossible under prior.
         """
         theta = torch.as_tensor(theta, dtype=torch.float32)
-        theta = ensure_theta_batched(theta)
-        num_batch = theta.shape[0]
-        x = ensure_x_batched(self.x).repeat(num_batch, 1)
-
-        with torch.set_grad_enabled(False):
-            # Evaluate on device, move back to cpu for comparison with prior.
-            log_likelihood = self.likelihood_nn.log_prob(
-                inputs=x, context=theta.to(x.device)
-            ).cpu()
 
         # Notice opposite sign to pyro potential.
-        return log_likelihood + self.prior.log_prob(theta)
+        return self.log_likelihood(theta).cpu() + self.prior.log_prob(theta)
 
-    def pyro_potential(self, theta: Dict[str, Tensor]) -> Tensor:
+    def pyro_potential(
+        self, theta: Dict[str, Tensor], track_gradients: bool = False
+    ) -> Tensor:
         r"""Return posterior log probability of parameters $p(\theta|x)$.
 
          Args:
@@ -365,9 +426,7 @@ class PotentialFunctionProvider:
 
         theta = next(iter(theta.values()))
 
-        # Evaluate on device, move back to cpu for comparison with prior.
-        log_likelihood = self.likelihood_nn.log_prob(
-            inputs=self.x.reshape(1, -1), context=theta.reshape(1, -1).to(self.x.device)
-        ).cpu()
-
-        return -(log_likelihood + self.prior.log_prob(theta))
+        return -(
+            self.log_likelihood(theta, track_gradients=track_gradients).cpu()
+            + self.prior.log_prob(theta)
+        )
