@@ -1,7 +1,6 @@
 # This file is part of sbi, a toolkit for simulation-based inference. sbi is licensed
 # under the Affero General Public License v3, see <https://www.gnu.org/licenses/>.
-
-
+import time
 from abc import ABC, abstractmethod
 from copy import deepcopy
 from typing import Any, Callable, Dict, Optional, Union
@@ -11,7 +10,6 @@ import torch
 from torch import Tensor, ones, optim
 from torch.nn.utils import clip_grad_norm_
 from torch.utils import data
-from torch.utils.data.sampler import SubsetRandomSampler
 from torch.utils.tensorboard import SummaryWriter
 
 from sbi import utils as utils
@@ -37,7 +35,7 @@ class PosteriorEstimator(NeuralInference, ABC):
         logging_level: Union[int, str] = "WARNING",
         summary_writer: Optional[SummaryWriter] = None,
         show_progress_bars: bool = True,
-        **unused_args
+        **unused_args,
     ):
         """Base class for Sequential Neural Posterior Estimation methods.
 
@@ -84,10 +82,7 @@ class PosteriorEstimator(NeuralInference, ABC):
         self._summary.update({"rejection_sampling_acceptance_rates": []})  # type:ignore
 
     def append_simulations(
-        self,
-        theta: Tensor,
-        x: Tensor,
-        proposal: Optional[Any] = None,
+        self, theta: Tensor, x: Tensor, proposal: Optional[Any] = None
     ) -> "PosteriorEstimator":
         r"""
         Store parameters and simulation outputs to use them for later training.
@@ -109,7 +104,7 @@ class PosteriorEstimator(NeuralInference, ABC):
             NeuralInference object (returned so that this function is chainable).
         """
 
-        validate_theta_and_x(theta, x)
+        theta, x = validate_theta_and_x(theta, x, training_device=self._device)
         self._check_proposal(proposal)
 
         if (
@@ -214,8 +209,11 @@ class PosteriorEstimator(NeuralInference, ABC):
         # Starting index for the training set (1 = discard round-0 samples).
         start_idx = int(discard_prior_samples and self._round > 0)
 
-        # For non-atomic loss, we can not reuse samples from prev. rounds as of now.
-        if self.use_non_atomic_loss:
+        # For non-atomic loss, we can not reuse samples from previous rounds as of now.
+        # SNPE-A can, by construction of the algorithm, only use samples from the last
+        # round. SNPE-A is the only algorithm that has an attribute `_ran_final_round`,
+        # so this is how we check for whether or not we are using SNPE-A.
+        if self.use_non_atomic_loss or hasattr(self, "_ran_final_round"):
             start_idx = self._round
 
         theta, x, prior_masks = self.get_simulations(
@@ -223,11 +221,7 @@ class PosteriorEstimator(NeuralInference, ABC):
         )
 
         # Dataset is shared for training and validation loaders.
-        dataset = data.TensorDataset(
-            theta,
-            x,
-            prior_masks,
-        )
+        dataset = data.TensorDataset(theta, x, prior_masks)
 
         # Set the proposal to the last proposal that was passed by the user. For
         # atomic SNPE, it does not matter what the proposal is. For non-atomic
@@ -252,6 +246,13 @@ class PosteriorEstimator(NeuralInference, ABC):
             self._neural_net = self._build_neural_net(
                 theta[self.train_indices], x[self.train_indices]
             )
+            # If data on training device already move net as well.
+            if (
+                not self._device == "cpu"
+                and f"{x.device.type}:{x.device.index}" == self._device
+            ):
+                self._neural_net.to(self._device)
+
             test_posterior_net_for_multi_d_x(self._neural_net, theta, x)
             self._x_shape = x_shape_from_simulation(x)
 
@@ -260,8 +261,7 @@ class PosteriorEstimator(NeuralInference, ABC):
 
         if not resume_training:
             self.optimizer = optim.Adam(
-                list(self._neural_net.parameters()),
-                lr=learning_rate,
+                list(self._neural_net.parameters()), lr=learning_rate
             )
             self.epoch, self._val_log_prob = 0, float("-Inf")
 
@@ -271,6 +271,8 @@ class PosteriorEstimator(NeuralInference, ABC):
 
             # Train for a single epoch.
             self._neural_net.train()
+            train_log_prob_sum = 0
+            epoch_start_time = time.time()
             for batch in train_loader:
                 self.optimizer.zero_grad()
                 # Get batches on current device.
@@ -282,26 +284,28 @@ class PosteriorEstimator(NeuralInference, ABC):
 
                 batch_loss = torch.mean(
                     self._loss(
-                        theta_batch,
-                        x_batch,
-                        masks_batch,
-                        proposal,
-                        calibration_kernel,
+                        theta_batch, x_batch, masks_batch, proposal, calibration_kernel
                     )
                 )
+
+                train_log_prob_sum += batch_loss.sum().item()
+
                 batch_loss.backward()
                 if clip_max_norm is not None:
                     clip_grad_norm_(
-                        self._neural_net.parameters(),
-                        max_norm=clip_max_norm,
+                        self._neural_net.parameters(), max_norm=clip_max_norm
                     )
                 self.optimizer.step()
 
             self.epoch += 1
 
+            train_log_prob_sum /= int(theta.shape[0] * (1.0 - validation_fraction))
+            self._summary["train_log_probs"].append(train_log_prob_sum)
+
             # Calculate validation performance.
             self._neural_net.eval()
             log_prob_sum = 0
+
             with torch.no_grad():
                 for batch in val_loader:
                     theta_batch, x_batch, masks_batch = (
@@ -311,11 +315,7 @@ class PosteriorEstimator(NeuralInference, ABC):
                     )
                     # Take negative loss here to get validation log_prob.
                     batch_log_prob = -self._loss(
-                        theta_batch,
-                        x_batch,
-                        masks_batch,
-                        proposal,
-                        calibration_kernel,
+                        theta_batch, x_batch, masks_batch, proposal, calibration_kernel
                     )
                     log_prob_sum += batch_log_prob.sum().item()
 
@@ -325,6 +325,7 @@ class PosteriorEstimator(NeuralInference, ABC):
             )
             # Log validation log prob for every epoch.
             self._summary["validation_log_probs"].append(self._val_log_prob)
+            self._summary["epoch_durations_sec"].append(time.time() - epoch_start_time)
 
             self._maybe_show_progress(self._show_progress_bars, self.epoch)
 
@@ -335,12 +336,7 @@ class PosteriorEstimator(NeuralInference, ABC):
         self._summary["best_validation_log_probs"].append(self._best_val_log_prob)
 
         # Update tensorboard and summary dict.
-        self._summarize(
-            round_=self._round,
-            x_o=None,
-            theta_bank=theta,
-            x_bank=x,
-        )
+        self._summarize(round_=self._round, x_o=None, theta_bank=theta, x_bank=x)
 
         # Update description for progress bar.
         if show_train_summary:
