@@ -1,4 +1,5 @@
-from typing import Callable, Any
+from typing import Any, Callable
+
 from pyro import poutine as poutine
 from torch.distributions import biject_to
 
@@ -10,12 +11,12 @@ def get_transforms(model: Callable, *model_args: Any, **model_kwargs: Any):
         model: Pyro model
         model_args: Arguments passed to model
         model_args: Keyword arguments passed to model
-    
+
     Example:
         ```python
         def prior():
             return pyro.sample("theta", pyro.distributions.Uniform(0., 1.))
-            
+
         transform_to_unbounded = get_transforms(prior)["theta"]
         ```
     """
