@@ -3,6 +3,7 @@
 
 import logging
 import warnings
+from math import pi
 from typing import Any, Dict, List, Optional, Sequence, Tuple, Union
 
 import torch
@@ -11,9 +12,12 @@ from pyro.distributions import Empirical
 from torch import Tensor, as_tensor
 from torch import nn as nn
 from torch import ones, zeros
+from torch.distributions import Independent
+from torch.distributions.distribution import Distribution
 from tqdm.auto import tqdm
 
-from sbi.utils.torchutils import BoxUniform, atleast_2d
+from sbi import utils as utils
+from sbi.utils.torchutils import atleast_2d
 
 
 def warn_if_zscoring_changes_data(x: Tensor, duplicate_tolerance: float = 0.1) -> None:
@@ -133,10 +137,10 @@ def standardizing_net(batch_t: Tensor, min_std: float = 1e-7) -> nn.Module:
     else:
         t_std = 1
         logging.warning(
-            f"""Using a one-dimensional batch will instantiate a Standardize transform 
-            with (mean, std) parameters which are not representative of the data. We allow
-            this behavior because you might be loading a pre-trained. If this is not the case, 
-            please be sure to use a larger batch."""
+            """Using a one-dimensional batch will instantiate a Standardize transform
+            with (mean, std) parameters which are not representative of the data. We
+            allow this behavior because you might be loading a pre-trained. If this is
+            not the case, please be sure to use a larger batch."""
         )
 
     return Standardize(t_mean, t_std)
@@ -306,20 +310,17 @@ def warn_on_invalid_x(num_nans: int, num_infs: int, exclude_invalid_x: bool) ->
             )
 
 
-def warn_on_invalid_x(num_nans: int, num_infs: int, exclude_invalid_x: bool) -> None:
-    """Warn if there are NaNs or Infs. Warning text depends on `exclude_invalid_x`."""
+def warn_on_iid_x(num_trials):
+    """Warn if more than one x was passed."""
 
-    if num_nans + num_infs > 0:
-        if exclude_invalid_x:
-            logging.warning(
-                f"Found {num_nans} NaN simulations and {num_infs} Inf simulations. "
-                "They will be excluded from training."
-            )
-        else:
-            logging.warning(
-                f"Found {num_nans} NaN simulations and {num_infs} Inf simulations. "
-                "Training might fail. Consider setting `exclude_invalid_x=True`."
-            )
+    if num_trials > 1:
+        warnings.warn(
+            f"An x with a batch size of {num_trials} was passed. "
+            + """It will be interpreted as a batch of independent and identically
+            distributed data X={x_1, ..., x_n}, i.e., data generated based on the
+            same underlying (unknown) parameter. The resulting posterior will be with
+            respect to entire batch, i.e,. p(theta | X)."""
+        )
 
 
 def warn_on_invalid_x_for_snpec_leakage(
@@ -330,14 +331,49 @@ def warn_on_invalid_x_for_snpec_leakage(
     if num_nans + num_infs > 0 and exclude_invalid_x:
         if algorithm == "SNPE_C" and round_ > 0:
             logging.warning(
-                f"When invalid simulations are excluded, multi-round SNPE-C"
-                f" can `leak` into the regions where parameters led to"
-                f" invalid simulations. This can lead to poor results."
+                "When invalid simulations are excluded, multi-round SNPE-C"
+                " can `leak` into the regions where parameters led to"
+                " invalid simulations. This can lead to poor results."
             )
 
 
+def check_warn_and_setstate(
+    state_dict: Dict, key_name: str, replacement_value: Any, warning_msg: str = ""
+) -> Tuple[Dict, str]:
+    """
+    Check if `key_name` is in `state_dict` and add it if not.
+
+    If the key already existed in the `state_dict`, the dictionary remains
+    unaltered. This function also appends to a warning string.
+
+    For developers: The reason that this method only appends to a warning string
+    instead of warning directly is that the user might get multiple very similar
+    warnings if multiple attributes had to be replaced. Thus, we start off with an
+    emtpy string and keep appending all missing attributes. Then, in the end,
+    all attributes are displayed along with a full description of the warning.
+
+    Args:
+        attribute_name: The name of the attribute to check.
+        state_dict: The dictionary to search (and write to if the key does not yet
+            exist).
+        replacement_value: The value to be written to the `state_dict`.
+        warning_msg: String to which the warning message should be appended to.
+
+    Returns:
+        A dictionary which contains the key `attribute_name` and a string with an
+        appended warning message.
+    """
+
+    if key_name not in state_dict.keys():
+        state_dict[key_name] = replacement_value
+        warning_msg += " `self." + key_name + f" = {str(replacement_value)}`"
+    return state_dict, warning_msg
+
+
 def get_simulations_since_round(
-    data: List, data_round_indices: List, starting_round_index: int
+    data: List,
+    data_round_indices: List,
+    starting_round_index: int,
 ) -> Tensor:
     """
     Returns tensor with all data coming from a round >= `starting_round`.
@@ -444,29 +480,41 @@ def logit(theta: Tensor, lower_bound: Tensor, upper_bound: Tensor) -> Tensor:
     return torch.log(theta_01 / (1 - theta_01))
 
 
-def check_if_boxuniform(dist) -> Tuple[bool, Optional[BoxUniform]]:
-    """Returns whether the `dist` is `BoxUniform` as well as the `BoxUniform` itself.
+def check_dist_class(
+    dist, class_to_check: Union[Distribution, Sequence[Distribution]]
+) -> Tuple[bool, Optional[Distribution]]:
+    """Returns whether the `dist` is instance of `class_to_check`.
 
-    When the user called `prepare_for_sbi`, the distribution will in fact be a
+    The dist can be hidden in an Independent distribution, a Boxuniform or in a wrapper.
+    E.g., when the user called `prepare_for_sbi`, the distribution will in fact be a
     `PytorchReturnTypeWrapper`. Thus, we need additional checks.
 
     Args:
         dist: Distribution to be checked.
 
     Returns:
-        Whether the `dist` is `BoxUniform` and the `BoxUniform` itself.
+        Whether the `dist` is `Uniform` and the `Uniform` itself.
     """
-    if isinstance(dist, BoxUniform):
-        is_boxuniform = True
-        box_dist = dist
-    elif hasattr(dist, "prior") and isinstance(dist.prior, BoxUniform):
-        is_boxuniform = True
-        box_dist = dist.prior
-    else:
-        is_boxuniform = False
-        box_dist = None
 
-    return is_boxuniform, box_dist
+    # Direct check.
+    if isinstance(dist, class_to_check):
+        return True, dist
+    # Reveal prior dist wrapped by user input checks or BoxUniform / Independent.
+    else:
+        if hasattr(dist, "prior"):
+            dist = dist.prior
+        if isinstance(dist, Independent):
+            dist = dist.base_dist
+
+        # Check dist.
+        if isinstance(dist, class_to_check):
+            is_instance = True
+            return_dist = dist
+        else:
+            is_instance = False
+            return_dist = None
+
+        return is_instance, return_dist
 
 
 def within_support(distribution: Any, samples: Tensor) -> Tensor:
@@ -478,29 +526,33 @@ def within_support(distribution: Any, samples: Tensor) -> Tensor:
     returns whether it is finite or not (this hanldes e.g. `NeuralPosterior`). Only
     checking whether the log-probabilty is not `-inf` will not work because, as of
     torch v1.8.0, a `torch.distribution` will throw an error at `log_prob()` when the
-    sample is out of the support (see #451). In `prepare_for_sbi()`, we set 
-    `validate_args=False`. Thish would take care of this, but requires running 
+    sample is out of the support (see #451). In `prepare_for_sbi()`, we set
+    `validate_args=False`. This would take care of this, but requires running
     `prepare_for_sbi()` and otherwise throws a cryptic error.
 
     Args:
-        distribution: Distribution under which to evaluate the `samples`.
+        distribution: Distribution under which to evaluate the `samples`, e.g., a
+            PyTorch distribution or NeuralPosterior.
         samples: Samples at which to evaluate.
 
     Returns:
         Tensor of bools indicating whether each sample was within the support.
     """
-    if hasattr(distribution, "support"):
+    # Try to check using the support property, use log prob method otherwise.
+    try:
         sample_check = distribution.support.check(samples)
 
-        # Before torch v1.7.0, `support.check()` returned bools for every element. From
-        # v1.8.0 on, it directly considers all dimensions of a sample. E.g., for a 
-        # single sample in 3D, v1.7.0 would return [[True, True, True]] and v1.8.0 
-        # would return [True].
+        # Before torch v1.7.0, `support.check()` returned bools for every element.
+        # From v1.8.0 on, it directly considers all dimensions of a sample. E.g.,
+        # for a single sample in 3D, v1.7.0 would return [[True, True, True]] and
+        # v1.8.0 would return [True].
         if sample_check.ndim > 1:
-            return torch.all(distribution.support.check(samples), axis=1)
+            return torch.all(sample_check, axis=1)
         else:
-            return distribution.support.check(samples)
-    else:
+            return sample_check
+    # Falling back to log prob method of either the NeuralPosterior's net, or of a
+    # custom wrapper distribution's.
+    except (NotImplementedError, AttributeError):
         return torch.isfinite(distribution.log_prob(samples))
 
 
@@ -529,3 +581,45 @@ class ImproperEmpirical(Empirical):
         """
         value = atleast_2d(value)
         return zeros(value.shape[0])
+
+
+def mog_log_prob(
+    theta: Tensor,
+    logits_pp: Tensor,
+    means_pp: Tensor,
+    precisions_pp: Tensor,
+) -> Tensor:
+    r"""
+    Returns the log-probability of parameter sets $\theta$ under a mixture of Gaussians.
+
+    Note that the mixture can have different logits, means, covariances for any theta in
+    the batch. This is because these values were computed from a batch of $x$ (and the
+    $x$ in the batch are not the same).
+
+    This code is similar to the code of mdn.py in pyknos, but it does not use
+    log(det(Cov)) = -2*sum(log(diag(L))), L being Cholesky of Precision. Instead, it
+    just computes log(det(Cov)). Also, it uses the above-defined helper
+    `_batched_vmv()`.
+
+    Args:
+        theta: Parameters at which to evaluate the mixture.
+        logits_pp: (Unnormalized) mixture components.
+        means_pp: Means of all mixture components. Shape
+            (batch_dim, num_components, theta_dim).
+        precisions_pp: Precisions of all mixtures. Shape
+            (batch_dim, num_components, theta_dim, theta_dim).
+
+    Returns: The log-probability.
+    """
+
+    _, _, output_dim = means_pp.size()
+    theta = theta.view(-1, 1, output_dim)
+
+    # Split up evaluation into parts.
+    weights = logits_pp - torch.logsumexp(logits_pp, dim=-1, keepdim=True)
+    constant = -(output_dim / 2.0) * torch.log(torch.tensor([2 * pi]))
+    log_det = 0.5 * torch.log(torch.det(precisions_pp))
+    theta_minus_mean = theta.expand_as(means_pp) - means_pp
+    exponent = -0.5 * utils.batched_mixture_vmv(precisions_pp, theta_minus_mean)
+
+    return torch.logsumexp(weights + constant + log_det + exponent, dim=-1)
