@@ -324,3 +324,10 @@ def batched_first_of_batch(t: Tensor) -> Tensor:
     Takes in a tensor of shape (N, M) and outputs tensor of shape (1,M).
     """
     return t[:1]
+
+
+def assert_all_finite(quantity: Tensor, description: str = "tensor") -> None:
+    """Raise if tensor quantity contains any NaN or Inf element."""
+
+    msg = f"NaN/Inf present in {description}."
+    assert torch.isfinite(quantity).all(), msg
