@@ -30,26 +30,26 @@ def test_api_snl_on_linearGaussian(num_dim: int, set_seed):
         num_dim: parameter dimension of the gaussian model
     """
     num_samples = 10
-    x_o = zeros(num_dim)
 
     prior_mean = zeros(num_dim)
     prior_cov = eye(num_dim)
     prior = MultivariateNormal(loc=prior_mean, covariance_matrix=prior_cov)
 
     simulator, prior = prepare_for_sbi(diagonal_linear_gaussian, prior)
-    inference = SNL(
-        prior,
-        show_progress_bars=False,
-    )
+    inference = SNL(prior, show_progress_bars=False)
 
     theta, x = simulate_for_sbi(simulator, prior, 1000, simulation_batch_size=50)
     _ = inference.append_simulations(theta, x).train(max_num_epochs=5)
-    posterior = inference.build_posterior().set_default_x(x_o)
 
-    posterior.sample(sample_shape=(num_samples,), x=x_o, mcmc_parameters={"thin": 3})
+    for num_trials in [1, 2]:
+        x_o = zeros((num_trials, num_dim))
+        posterior = inference.build_posterior().set_default_x(x_o)
+        posterior.sample(
+            sample_shape=(num_samples,), x=x_o, mcmc_parameters={"thin": 3}
+        )
 
 
-def test_c2st_snl_on_linearGaussian_different_dims(set_seed):
+def test_c2st_snl_on_linearGaussian(set_seed):
     """Test whether SNL infers well a simple example with available ground truth.
 
     This example has different number of parameters theta than number of x. This test
@@ -75,7 +75,7 @@ def test_c2st_snl_on_linearGaussian_different_dims(set_seed):
     prior_cov = eye(theta_dim)
     prior = MultivariateNormal(loc=prior_mean, covariance_matrix=prior_cov)
     target_samples = samples_true_posterior_linear_gaussian_mvn_prior_different_dims(
-        x_o[0],
+        x_o,
         likelihood_shift,
         likelihood_cov,
         prior_mean,
@@ -84,15 +84,13 @@ def test_c2st_snl_on_linearGaussian_different_dims(set_seed):
         num_samples=num_samples,
     )
 
-    simulator = lambda theta: linear_gaussian(
-        theta, likelihood_shift, likelihood_cov, num_discarded_dims=discard_dims
-    )
-
-    simulator, prior = prepare_for_sbi(simulator, prior)
-    inference = SNL(
+    simulator, prior = prepare_for_sbi(
+        lambda theta: linear_gaussian(
+            theta, likelihood_shift, likelihood_cov, num_discarded_dims=discard_dims
+        ),
         prior,
-        show_progress_bars=False,
     )
+    inference = SNL(prior, show_progress_bars=False)
 
     theta, x = simulate_for_sbi(
         simulator, prior, num_simulations, simulation_batch_size=50
@@ -108,7 +106,9 @@ def test_c2st_snl_on_linearGaussian_different_dims(set_seed):
 @pytest.mark.slow
 @pytest.mark.parametrize("num_dim", (1, 2))
 @pytest.mark.parametrize("prior_str", ("uniform", "gaussian"))
-def test_c2st_snl_on_linearGaussian(num_dim: int, prior_str: str, set_seed):
+def test_c2st_snl_on_linearGaussian_different_dims_and_trials(
+    num_dim: int, prior_str: str, set_seed
+):
     """Test SNL on linear Gaussian, comparing to ground truth posterior via c2st.
 
     Args:
@@ -116,63 +116,80 @@ def test_c2st_snl_on_linearGaussian(num_dim: int, prior_str: str, set_seed):
         prior_str: one of "gaussian" or "uniform"
         set_seed: fixture for manual seeding
     """
-
-    x_o = zeros((1, num_dim))
     num_samples = 500
+    num_simulations = 7500
+    trials_to_test = [1, 5, 10]
 
     # likelihood_mean will be likelihood_shift+theta
     likelihood_shift = -1.0 * ones(num_dim)
-    likelihood_cov = 0.3 * eye(num_dim)
+    # Use increased cov to avoid too small posterior cov for many trials.
+    likelihood_cov = 0.8 * eye(num_dim)
 
     if prior_str == "gaussian":
         prior_mean = zeros(num_dim)
         prior_cov = eye(num_dim)
         prior = MultivariateNormal(loc=prior_mean, covariance_matrix=prior_cov)
-        gt_posterior = true_posterior_linear_gaussian_mvn_prior(
-            x_o[0], likelihood_shift, likelihood_cov, prior_mean, prior_cov
-        )
-        target_samples = gt_posterior.sample((num_samples,))
     else:
         prior = utils.BoxUniform(-2.0 * ones(num_dim), 2.0 * ones(num_dim))
-        target_samples = samples_true_posterior_linear_gaussian_uniform_prior(
-            x_o, likelihood_shift, likelihood_cov, prior=prior, num_samples=num_samples
+
+    simulator, prior = prepare_for_sbi(
+        lambda theta: linear_gaussian(theta, likelihood_shift, likelihood_cov), prior
+    )
+    inference = SNL(prior, show_progress_bars=False)
+
+    theta, x = simulate_for_sbi(
+        simulator, prior, num_simulations, simulation_batch_size=50
+    )
+    _ = inference.append_simulations(theta, x).train()
+
+    # Test inference amortized over trials.
+    for num_trials in trials_to_test:
+        x_o = zeros((num_trials, num_dim))
+        if prior_str == "gaussian":
+            gt_posterior = true_posterior_linear_gaussian_mvn_prior(
+                x_o, likelihood_shift, likelihood_cov, prior_mean, prior_cov
+            )
+            target_samples = gt_posterior.sample((num_samples,))
+        else:
+            target_samples = samples_true_posterior_linear_gaussian_uniform_prior(
+                x_o,
+                likelihood_shift,
+                likelihood_cov,
+                prior=prior,
+                num_samples=num_samples,
+            )
+        posterior = inference.build_posterior(
+            mcmc_method="slice_np_vectorized"
+        ).set_default_x(x_o)
+
+        samples = posterior.sample(
+            sample_shape=(num_samples,), mcmc_parameters={"thin": 3, "num_chains": 2}
         )
 
-    simulator = lambda theta: linear_gaussian(theta, likelihood_shift, likelihood_cov)
+        # Check performance based on c2st accuracy.
+        check_c2st(
+            samples, target_samples, alg=f"snle_a-{prior_str}-prior-{num_trials}-trials"
+        )
 
-    simulator, prior = prepare_for_sbi(simulator, prior)
-    inference = SNL(
-        prior,
-        show_progress_bars=False,
-    )
+        map_ = posterior.map(num_init_samples=1_000, init_method="prior")
 
-    theta, x = simulate_for_sbi(simulator, prior, 1000, simulation_batch_size=50)
-    _ = inference.append_simulations(theta, x).train()
-    posterior = inference.build_posterior().set_default_x(x_o)
+        # TODO: we do not have a test for SNL log_prob(). This is because the output
+        # TODO: density is not normalized, so KLd does not make sense.
+        if prior_str == "uniform":
+            # Check whether the returned probability outside of the support is zero.
+            posterior_prob = get_prob_outside_uniform_prior(posterior, prior, num_dim)
+            assert (
+                posterior_prob == 0.0
+            ), "The posterior probability outside of the prior support is not zero"
 
-    samples = posterior.sample(sample_shape=(num_samples,), mcmc_parameters={"thin": 3})
-
-    # Check performance based on c2st accuracy.
-    check_c2st(samples, target_samples, alg=f"snle_a-{prior_str}-prior")
-
-    map_ = posterior.map(num_init_samples=1_000, init_method="prior")
-
-    # TODO: we do not have a test for SNL log_prob(). This is because the output
-    # TODO: density is not normalized, so KLd does not make sense.
-    if prior_str == "uniform":
-        # Check whether the returned probability outside of the support is zero.
-        posterior_prob = get_prob_outside_uniform_prior(posterior, num_dim)
-        assert (
-            posterior_prob == 0.0
-        ), "The posterior probability outside of the prior support is not zero"
-
-        assert ((map_ - ones(num_dim)) ** 2).sum() < 0.5
-    else:
-        assert ((map_ - gt_posterior.mean) ** 2).sum() < 0.5
+            assert ((map_ - ones(num_dim)) ** 2).sum() < 0.5
+        else:
+            assert ((map_ - gt_posterior.mean) ** 2).sum() < 0.5
 
 
 @pytest.mark.slow
-def test_c2st_multi_round_snl_on_linearGaussian(set_seed):
+@pytest.mark.parametrize("num_trials", (1, 5))
+def test_c2st_multi_round_snl_on_linearGaussian(num_trials: int, set_seed):
     """Test SNL on linear Gaussian, comparing to ground truth posterior via c2st.
 
     Args:
@@ -180,8 +197,9 @@ def test_c2st_multi_round_snl_on_linearGaussian(set_seed):
     """
 
     num_dim = 2
-    x_o = zeros((1, num_dim))
+    x_o = zeros((num_trials, num_dim))
     num_samples = 500
+    num_simulations_per_round = 1000 * num_trials
 
     # likelihood_mean will be likelihood_shift+theta
     likelihood_shift = -1.0 * ones(num_dim)
@@ -191,25 +209,26 @@ def test_c2st_multi_round_snl_on_linearGaussian(set_seed):
     prior_cov = eye(num_dim)
     prior = MultivariateNormal(loc=prior_mean, covariance_matrix=prior_cov)
     gt_posterior = true_posterior_linear_gaussian_mvn_prior(
-        x_o[0], likelihood_shift, likelihood_cov, prior_mean, prior_cov
+        x_o, likelihood_shift, likelihood_cov, prior_mean, prior_cov
     )
     target_samples = gt_posterior.sample((num_samples,))
 
-    simulator = lambda theta: linear_gaussian(theta, likelihood_shift, likelihood_cov)
-
-    simulator, prior = prepare_for_sbi(simulator, prior)
-    inference = SNL(
-        prior,
-        show_progress_bars=False,
+    simulator, prior = prepare_for_sbi(
+        lambda theta: linear_gaussian(theta, likelihood_shift, likelihood_cov), prior
     )
+    inference = SNL(prior, show_progress_bars=False)
 
-    theta, x = simulate_for_sbi(simulator, prior, 750, simulation_batch_size=50)
+    theta, x = simulate_for_sbi(
+        simulator, prior, num_simulations_per_round, simulation_batch_size=50
+    )
     _ = inference.append_simulations(theta, x).train()
     posterior1 = inference.build_posterior(
         mcmc_method="slice_np_vectorized", mcmc_parameters={"thin": 5, "num_chains": 20}
     ).set_default_x(x_o)
 
-    theta, x = simulate_for_sbi(simulator, posterior1, 750, simulation_batch_size=50)
+    theta, x = simulate_for_sbi(
+        simulator, posterior1, num_simulations_per_round, simulation_batch_size=50
+    )
     _ = inference.append_simulations(theta, x).train()
     posterior = inference.build_posterior().copy_hyperparameters_from(posterior1)
 
@@ -222,10 +241,7 @@ def test_c2st_multi_round_snl_on_linearGaussian(set_seed):
 @pytest.mark.slow
 @pytest.mark.parametrize(
     "mcmc_method, prior_str",
-    (
-        ("slice", "gaussian"),
-        ("slice", "uniform"),
-    ),
+    (("slice", "gaussian"), ("slice", "uniform")),
 )
 def test_api_snl_sampling_methods(mcmc_method: str, prior_str: str, set_seed):
     """Runs SNL on linear Gaussian and tests sampling from posterior via mcmc.
@@ -246,10 +262,7 @@ def test_api_snl_sampling_methods(mcmc_method: str, prior_str: str, set_seed):
         prior = utils.BoxUniform(-1.0 * ones(num_dim), ones(num_dim))
 
     simulator, prior = prepare_for_sbi(diagonal_linear_gaussian, prior)
-    inference = SNL(
-        prior,
-        show_progress_bars=False,
-    )
+    inference = SNL(prior, show_progress_bars=False)
 
     theta, x = simulate_for_sbi(simulator, prior, 200, simulation_batch_size=50)
     _ = inference.append_simulations(theta, x).train(max_num_epochs=5)
